/*
    WOCFS - Write-Only enCrypted File System
    Copyright (C) 2011, Michael Rao

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
     any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "utils.hpp"
#include <zlib.h>

int log=1;

struct mainpart_t {
  ll_t t;
  ll_t date;
  unsigned char hash[AESKEYSIZE];

  unsigned char hash2[AESKEYSIZE];
  ll_t i2;
  void inithash2() {
    memcpy(hash2,hash,AESKEYSIZE);
    i2=0;
  }
  const unsigned char *gethash(ll_t i)
  {
    if(i<i2)
      inithash2();
    while(i2<i) {
      unsigned char tp[AESKEYSIZE];
      gcry_md_hash_buffer(GCRY_MD_SHA256,tp,hash2,AESKEYSIZE);
      memcpy(hash2,tp,AESKEYSIZE);
      i2++;
    }
    return hash2;
  }
};

vector<mainpart_t> mainparts;
map<ll_t,ll_t> o2i;

int main(int ac, char **av)
{
  while(1) {
    if(ac<2) break;
    if(0==strcmp(av[1],"-v")) { log++; ac--;av++;continue;}
    if(0==strcmp(av[1],"-q")) { log--; ac--;av++;continue;}
    break;
  }

  gcry_control( GCRYCTL_DISABLE_SECMEM );
  const char *version = gcry_check_version(NULL);
  if(log>1) fprintf(stderr,"Libgcrypt version: %s\n",version);
  
  gcry_sexp_t skey;
  read_sec_key(&skey, "wocfs.sec_key");

  char check[BLOCKSIZE];
  
  {
    FILE *in=fopen("wocfs.check","r");
    if(in==NULL) {
      if(log>0) fprintf(stderr,"WARN: No checkfile !!!\n");
    } else {
      int re=fread(check,1,BLOCKSIZE,in);
      if(re!=BLOCKSIZE) {
	if(log>0) fprintf(stderr,"WARN: bad checkfile\n");
      }
    }
  }
  
  FILE *in=fopen("wocfs.main","r");
  while(!feof(in)) {

    int si;
    int re=fread(&si,1,4,in);
    if(re!=4) {
      if(re==0) break;
      if(log>0) fprintf(stderr,"WARN: incomplete mainfile\n");
      break;
    }

    gcry_sexp_t ctexp;
    {
      char bf[si];
      re=fread(bf,1,si,in);
      if(re!=si) {
	if(log>0) fprintf(stderr,"WARN: incomplete mainfile\n");
	break;
      }
      
      int r=gcry_sexp_build(&ctexp,NULL,"(enc-val (rsa (a %b)))",si,bf); // (flags pkcs1) 
      if(r) {
	fprintf(stderr,"Error in gcry_sexp_build\n");
	exit(1);
      }
      memset(bf,0,si);
    }

    gcry_sexp_t texp;
    int r=gcry_pk_decrypt(&texp, ctexp, skey);
    if(r) {
      fprintf(stderr,"Error in gcry_pk_decrypt\n");
      exit(1);
    }
    gcry_sexp_release(ctexp);

    /*
    {
      char bf[100000];
      r=gcry_sexp_sprint(texp,GCRYSEXP_FMT_ADVANCED,bf,100000);
      assert(r>0 && r<100000);
      printf("%s",bf);
    }
    */

    size_t len;
    const unsigned char *data;
    data =(unsigned char *) gcry_sexp_nth_data (texp, 0, &len);

    unsigned char data2[AESKEYSIZE+48];
    memset(data2,0,AESKEYSIZE+48);

    {
      unsigned int i;
      for(i=1;i<len-AESKEYSIZE+48;i++) {
	if(data[i]==0) {
	  memcpy(data2,data+i+1,AESKEYSIZE+48);
	  i=0;
	  break;
	}
      }
      if(i) {
	if(log>0) fprintf(stderr,"ERROR in mainfile !\n");
	break;
      }
    }

    /*
    for(int i=0;i<32+24;i++) {
      printf("%02X",data2[i]);
    }
    printf("\n");
    */

    if(memcmp(data2,M_HEAD,4)) {
      if(log>0) fprintf(stderr,"ERROR in mainfile (head) !\n");
      break;
    }

    int flags=readint(data2+4,4);
    if(flags != 1+M_DATE+M_MAJOR+M_HASH 
       && flags != M_CONF+M_DATE+M_MAJOR+M_HASH) {
      if(log>0) fprintf(stderr,"ERROR in mainfile (flags) !\n");
      break;
    }
    
    int d=8;
    
    if(flags& M_CONF) {
      d+=4;
      int f2=readint(data2+d,4); d+=4;
      if(f2) {
	if(log>0) fprintf(stderr,"ERROR in mainfile (flags2) !\n");
	break;
      }
    }

    mainpart_t m;
    m.t=(flags&1);
    
    if(flags&M_DATE) {
      m.date=readint(data2+d,8);d+=8;
    }
    ll_t maj;
    if(flags&M_MAJOR) {
      maj=readint(data2+d,8);d+=8;
    }
    //printf("maj=%Ld\n",maj);

    if(flags&M_HASH) {
      memcpy(&(m.hash),data2+d,AESKEYSIZE);d+=AESKEYSIZE;
    }
    
    if(m.t==0) {
      if(o2i.find(maj)!=o2i.end()) {
	if(log>0) fprintf(stderr,"Warn: duplicate Major number %Ld\n",maj);
      }
      o2i[maj]=mainparts.size();
    }
    
    mainparts.resize(mainparts.size()+1);
    mainparts[mainparts.size()-1]=m;
    mainparts[mainparts.size()-1].inithash2();
    
    char ct[30];
    time_t tim;
    tim=m.date/1000;
    ctime_r(&tim,ct);
    if(m.t==0) {
      if(log>1) fprintf(stderr,"start : ");
    }
    else if(m.t==1) {
      if(log>1) fprintf(stderr,"stop  : ");
    }
    else {
      if(log>0) fprintf(stderr,"ERROR: unkwown starting type of mainblock\n");
      if(log>1) fprintf(stderr,"error : ");
      //break;
    }
    if(log>1) {

      /*
      for(int i=0;i<AESKEYSIZE;i++) {
	fprintf(stderr,"%02X",int(m.hash[i]));
      }
      */
      
      fprintf(stderr," %s",ct);
    }
  }

  if(mainparts.size()==0) {
    if(memcmp(check,"0123456789ABCDEF0123456789ABCDEF",BLOCKSIZE)==0) {
      if(log>1) fprintf(stderr,"check OK (main is empty)\n");
    }
  } else {
    int i;
    for(i=mainparts.size()-1;i>=0;i--) {
      gcry_error_t r;
      gcry_cipher_hd_t hd;
      r=gcry_cipher_open (&hd, GCRY_CIPHER_AES256, GCRY_CIPHER_MODE_ECB,0 & GCRY_CIPHER_SECURE);
      assert(r==0);
      r=gcry_cipher_setkey(hd,mainparts[i].hash,AESKEYSIZE);
      assert(r==0);
      char ch2[BLOCKSIZE];
      r=gcry_cipher_decrypt(hd, ch2,BLOCKSIZE,check,BLOCKSIZE);
      assert(r==0);
      memcpy(check,ch2,BLOCKSIZE);
      gcry_cipher_close(hd);
      if(memcmp(check,"0123456789ABCDEF0123456789ABCDEF",BLOCKSIZE)==0) {
	if(i!=0) { 
	  if(log>0) fprintf(stderr,"WARN: check newer than main !\n");
	} else {
	  if(log>1) fprintf(stderr,"check OK\n");
	}
	break;
      }
    }
    if(i<0) {
      if(log>0) fprintf(stderr,"Warn: wrong check !\n");
    }
  }
  
  for(int i=1;i<ac;i++) {
    FILE *in=fopen(av[i],"r");
    if(in==NULL) {
      if(log>0) fprintf(stderr,"Cannot open %s\n",av[i]);
      continue;
    }

    bool first=1;
    ll_t oma=0,omi=0;
    while(!feof(in)) {
      unsigned char bfs[4];
      int r=fread(bfs,1,4,in);
      if(r<4) {
	if(r) if(log>0) fprintf(stderr,"incomplete last block (%d)\n",r);
	break;
      }
      ll_t si=readint(bfs,4);
      if(si<=0 || si>(1LL<<30)) {
	if(log>0) fprintf(stderr,"invalid size %Ld\n",si);
	break;
      }
      unsigned char *bf=(unsigned char*)malloc(si);
      r=fread(bf,1,si,in);
      if(r<si) {
	if(log>0) fprintf(stderr,"incomplete last block (%d<%Ld)\n",r,si);
	break; 
      }

      ll_t ma,mi;
      int xx=0;
      xx+=readcint(bf+xx,ma);
      xx+=readcint(bf+xx,mi);
      if((si-xx)%BLOCKSIZE) {
	if(log>0) fprintf(stderr,"incomplete last block (%%BLOCKSIZE=%Ld)\n",(si-xx)%BLOCKSIZE);
	break;
      }
      unsigned char *bfd=(unsigned char*)malloc(si-xx);

      map<ll_t,ll_t>::iterator it=o2i.find(ma);
      if(it==o2i.end()) {
	if(log>0) fprintf(stderr,"unknown major %Ld\n",ma);
	break;
      }
      int mai=it->second;

      gcry_error_t rc;
      gcry_cipher_hd_t hd;
      rc=gcry_cipher_open (&hd, GCRY_CIPHER_AES256, GCRY_CIPHER_MODE_CBC,0 & GCRY_CIPHER_SECURE);
      assert(rc==0);
      const unsigned char *key=mainparts[mai].gethash(mi);

      /*
      for(int i=0;i<AESKEYSIZE;i++) {
	printf("%02X",int(key[i]));
      }
      printf(" << KEY for %Ld (%d) %Ld\n",ma,mai,mi);
      */

      rc=gcry_cipher_setkey(hd,key,AESKEYSIZE);
      assert(rc==0);
      rc=gcry_cipher_decrypt(hd,bfd,si-xx,bf+xx,si-xx);
      assert(rc==0);
      gcry_cipher_close(hd);

      /*
      for(int i=0;i<20;i++) {
	printf("%02X",int(bfd[i]));
      }
      printf("\n");
      */

      ll_t flags;
      flags=readint(bfd,4);
      int yy=4;
      ll_t oma_,omi_=0;
      if(flags&BF_PREV_MAJ)
	yy+=readcint(bfd+yy,oma_);
      else oma_=ma;
      if(flags&BF_PREV_MIN)
	yy+=readcint(bfd+yy,omi_);
      ll_t bsize;
      yy+=readcint(bfd+yy,bsize);
      
      //verif ma,mi
      if(oma_==0 && omi_==0) {
	if(!first) {
	  if(log>0) fprintf(stderr,"WARN: indicate a starting block\n");
	} else {
	  if(log>2) fprintf(stderr,"<<starting block>>");
	}
      } else if(oma_==1 && omi_==0) {
	if(first) {
	  if(log>0) fprintf(stderr,"WARN: indicate a continuing block\n");
	} else {
	  if(log>1) fprintf(stderr,"<<continuing block>>");
	}
      } else {
	if(oma_!=oma || omi!=omi) {
	  if(log>0) fprintf(stderr,"WARN: bad previous block (%Ld %Ld)!=(%Ld %Ld)\n",oma_,omi_,oma,omi);
	} else {
	  if(log>2) fprintf(stderr,"<<good previous block>>");
	}
      }

      if(bsize<=0 || bsize+yy>si-xx) {
	if(log>0) fprintf(stderr,"ERROR: over/under-full bsize %Ld\n",bsize);
      } else {
	if(log>2) fprintf(stderr,"<<block size %Ld>>",bsize);
	int t;
	/*
	for(t=yy+bsize;t<si-xx;t++)
	  printf("%02X",bfd[t]);
	printf("\n");*/
	for(t=yy+bsize;t<si-xx;t++)
	  if(bfd[t]) break;
	if(t<si-xx) {
	  if(log>0) fprintf(stderr,"ERROR: not filled with zeros\n");
	}
	if(flags&BF_COMP) {
	  if(log>2) fprintf(stderr,"<<compressed>>");
	  unsigned char *bfc=bfd+yy;
	  unsigned char *bfu=(unsigned char *)malloc(bsize*2);
	  ll_t bfusize=bsize*2;
	  z_stream strm;
	  int r;
	  strm.zalloc=Z_NULL;
	  strm.zfree=Z_NULL;
	  strm.opaque=Z_NULL;    
	  strm.avail_in = 0;
	  strm.next_in = Z_NULL;
	  if(Z_OK!=inflateInit(&strm)) {
	    if(log>0) fprintf(stderr,"Error in inflateInit\n");
	    goto fail_decomp;
	  }
	  strm.next_in=(Bytef*)bfc;
	  strm.avail_in=bsize;
	  do {
	    strm.next_out=(Bytef*)bfu;
	    strm.avail_out=bfusize;
	    r=inflate(&strm,Z_NO_FLUSH);
	    if(r==Z_STREAM_ERROR) {
	      if(log>0) fprintf(stderr,"Error: inflate returns STREAM_ERROR\n");
	      goto fail_decomp;
	    }
	    if(r==Z_DATA_ERROR) {
	      if(log>1) fprintf(stderr,"Error: inflate returns DATA_ERROR\n");
	      goto fail_decomp;
	    }
	    if(r==Z_NEED_DICT) {
	      if(log>0) fprintf(stderr,"Error: inflate returns NEED_DICT\n");
	      goto fail_decomp;
	    }
	    fwrite(bfu,1,bfusize-strm.avail_out,stdout);
	  } while(strm.avail_out==0);
	  inflateEnd(&strm);
	fail_decomp:
	  free(bfu);
	} else {
	  //assert(flags==0);
	  fwrite(bfd+yy,1,bsize,stdout);
	}
      }

      free(bf);
      free(bfd);
      
      oma=ma;omi=mi;
      first=0;
    }

    fclose(in);
  }
  
  gcry_sexp_release( skey );

  return 0;
}
