/*
    WOCFS - Write-Only enCrypted File System
    Copyright (C) 2011, Michael Rao

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
     any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <utime.h>
#include <sys/time.h>
#include <unistd.h>
#include <dirent.h>
#include <assert.h>
#include <sys/stat.h>
#include <time.h>

#include <gcrypt.h>

#include <string>
#include <vector>
#include <map>
using namespace std;

typedef long long ll_t;
typedef unsigned long long ull_t;

#define BLOCKSIZE 16
#define AESKEYSIZE 32

#include <termios.h>

struct termios oldt, newt;
bool echo_off() {
  if(tcgetattr(fileno(stdin),&oldt)!=0)
    return 0;
  newt=oldt;
  newt.c_lflag&=~ECHO||ECHOCTL;
  if(tcsetattr (fileno(stdin),TCSAFLUSH,&newt)!=0)
    return 0;
  return 1;
}

bool echo_on() {
  if(tcsetattr(fileno(stdin),TCSAFLUSH,&oldt)!=0) 
    return 0;
  return 1;
}  

bool getpass(char *pass,int ma)
{
  int i=0;
  
  echo_off();
  while((pass[i]=getchar())!='\n') {
    i++;
    if(i==ma) return false;
  }
  echo_on();
  pass[i]=0;
  fprintf(stderr,"\n");
  return true;
}

void read_sec_key(gcry_sexp_t *skey,const char *fn, bool hashed=1)
{
  int rc;

  char bf[100000];
  char bc[100000];
  char pass[10000];
  char hash[32];

  if(hashed) {
    fprintf(stderr,"pass: ");
    getpass(pass,10000);
    
    gcry_md_hash_buffer(GCRY_MD_SHA256,hash,pass,strlen(pass));
  }

  FILE *in=fopen(fn,"r");
  assert(in);
  int b=fread(bc,1,100000,in);
  //bf[b]=0;
  assert(b%32==0);
  fclose(in);
  
  if(hashed) {
    gcry_error_t r;
    gcry_cipher_hd_t hd;
    r=gcry_cipher_open (&hd, GCRY_CIPHER_AES256, GCRY_CIPHER_MODE_CBC,GCRY_CIPHER_SECURE);
    assert(r==0);
    r=gcry_cipher_setkey(hd,hash,32);
    assert(r==0);
    r=gcry_cipher_decrypt(hd, bf,b,bc,b);
    bf[b]=0;
    assert(r==0);
    gcry_cipher_close(hd);
  }
  
  rc = gcry_sexp_new( skey, bf, 0, 1 );
  
  bzero(bf,sizeof(bf));
  bzero(bc,sizeof(bc));
}

bool read_pub_key(const char *fn,gcry_sexp_t *pkey)
{
  int rc;
  char bf[100000];

  {
    FILE *in=fopen(fn,"r");
    if(in==0) return false;
    int b=fread(bf,1,100000,in);
    bf[b]=0;
    rc = gcry_sexp_new( pkey, bf, 0, 1 );
    if(rc) return false;
    fclose(in);
  }
  memset(bf,0,sizeof(bf));
  return 1;
}



bool writeint(unsigned char *bf,ll_t a,int n)
{
  for(int i=n-1;i>=0;i--) {
    bf[n-1-i]=((unsigned char*)&a)[i];
  }
  return true;
}

ll_t readint(const unsigned char *bf,int n)
{
  ll_t r=0;
  if(bf[0]&128) {
    r=-1;
  }
  for(int i=n-1;i>=0;i--) {
    r<<=8;
    r+=*bf;
    bf++;
  }
  return r;
}

//TODO:
int writecint(unsigned char *bf,ll_t i)
{
  writeint(bf,i,8);
  return 8;
}

int readcint(const unsigned char *bf,ll_t &i)
{
  i=readint(bf,8);
  return 8;
}

/*begin of main block: 8 bytes */
#define M_HEAD "WOFS"
#define M_TYPE (1+2) //0:start 1:stop...
#define M_CONF 4
#define M_DATE 8
#define M_MAJOR 16
#define M_MINOR 32
#define M_HASHID 256
#define M_HASH 512

// "conf_block" in "start block" 
// version: 4bytes
#define M_VER_0 0
#define M_VER_1 0
#define M_VER_2 1
#define M_VER_3 0
// general: 4 bytes ++
#define MF_EXTFLAG 1 //0
#define MF_NU1 2 //0
#define MF_INT // 2b
#define MF_GZ  // 2b
#define M_DICT //default dict
//#define MF_REC // 2b // useless

/* block begin: 4 bytes*/
#define BF_EXTFLAG 1 //0
#define BF_NC1 2 //0
#define BF_PREV_MAJ 4 //0/1
#define BF_PREV_MIN 8 //1 
#define BF_STARTDATE 256 // 0, 1, 2:diff
#define BF_ENDDATE 1024 // 0, 1, 2:diff
#define BF_ATOM 4096

#define BF_DICT (1<<17) //0

#define BF_COMP (1<<16)


